#ifndef QUAD_H_
#define QUAD_H_

#define	QUAD(klass)		((struct quad *)(klass))
#define	QUAD_ARG(klass)		((struct quad_arg *)(klass))
#define	QUAD_ID(klass)		((struct quad_id *)(klass))
#define	QUAD_REFER(klass)	((struct quad_refer *)(klass))
#define	QUAD_CONST(klass)	((struct quad_const *)(klass))

typedef enum quadruple_type {
	QUAD_UNKNOWN,
	QUAD_ARG,
	QUAD_CONST,
	QUAD_ID,
	QUAD_REFER,
	QUAD_LABEL,
	QUAD_GOTO,
	QUAD_IF,
	QUAD_CALL,
	QUAD_PARAM,
	QUAD_OP,
	QUAD_RETURN,
	QUAD_ASSIGN,
	QUAD_STORE_ACCESS,
	QUAD_LOAD_ACCESS,
#ifdef _MEM_AREA
	QUAD_STORE_MEM_AREA,
	QUAD_LOAD_MEM_AREA,
#endif
	QUAD_FUNC,
	QUAD_SCOPE,
	QUAD_SCOPE_END,
	QUAD_FUNC_END
} quadruple_type_t;

struct quad;
typedef void  (*quad_release_t)(struct quad *);

struct quad {
	quadruple_type_t quad_type;
	quad_release_t	release;
	unsigned int	ref_count;
};

struct quad_arg {
	struct quad base;
	data_type_t data_type;
};

struct quad_id {
	struct quad_arg base;
	char *id;
	boolean_t temp;
};

struct quad_refer {
	struct quad_arg base;
	char *id;
	boolean_t temp;
};

struct quad_const {
	struct quad_arg	base;
	union {
		int	integer;
		float	real;
		char	ch;
		char	*str;
	};
};

struct quad_label {
	struct quad base;
	int ln;
};

struct quad_goto {
	struct quad base;
	int ln;
};

struct quad_if {
	struct quad base;
	opcode_t opcode;
	struct quad *left;
	struct quad *right;
	int ln;
};

struct q_call_arg {
	data_type_t arg_type;
	boolean_t   is_ref;
};

struct quad_call {
	struct quad base;
	struct quad *ret_val;
	char *name;
	struct q_call_arg *args;
	unsigned int nargs;
	data_type_t ret_type;
};

struct quad_param {
	struct quad base;
	struct quad *param;
	boolean_t   is_ref;
};

struct quad_op {
	struct quad base;
	opcode_t opcode;
	struct quad *result;
	struct quad *left;
	struct quad *right;
};

struct quad_func {
	struct quad base;
	char *name;
	data_type_t ret_type;
	struct q_func_arg *args;
	unsigned int nargs;
};

struct quad_assign {
	struct quad base;
	struct quad *lvalue;
	struct quad *rvalue;
	boolean_t   leave;
};

struct quad_scope {
	struct quad base;
	struct symbol_table *table;
};

struct quad_scope_end {
	struct quad base;
};

struct quad_func_end {
	struct quad base;
};

struct quad_store_access {
	struct quad base;
	char *name;
	struct quad *idx;
	struct quad *rvalue;
	boolean_t   leave;
};

struct quad_load_access {
	struct quad base;
	struct quad *lvalue;
	char *name;
	struct quad *idx;
};

#ifdef _MEM_AREA
struct quad_store_mem_area {
	struct quad base;
	char	area;
	char	mem_data_type;
	int	offset;
	int	nbits;
	struct quad *idx;
	struct quad *rvalue;
	boolean_t is_unsign;
	boolean_t leave;
};

struct quad_load_mem_area {
	struct quad base;
	struct quad *lvalue;
	char	area;
	char	mem_data_type;
	int	offset;
	int	nbits;
	struct quad *idx;
	boolean_t is_unsign;
};
#endif

struct quad_const	*quad_const(data_type_t data_type, void *data);
struct quad_id		*quad_id(data_type_t decl_type, const char *name);
struct quad_refer	*quad_refer(data_type_t decl_type, const char *name);
struct quad_if		*quad_if(opcode_t opcode, struct quad *left, struct quad *right, int ln);
struct quad_goto	*quad_goto(int ln);
struct quad_label	*quad_label(int ln);
struct quad_call	*quad_call(const char *name, struct quad *ret_val, data_type_t ret_type);
void quad_call_add_arg(struct quad_call *call_quad, data_type_t arg_type, boolean_t is_ref);

struct quad_param	*quad_param(struct quad *param, boolean_t is_ref);
struct quad_func	*quad_func(const char *name, data_type_t ret_type);
struct quad_assign	*quad_assign(struct quad *lvalue, struct quad *rvalue);
struct quad_op		*quad_op(opcode_t opcode, struct quad *result, struct quad *left, struct quad *right);
struct quad_scope	*quad_scope(struct symbol_table *table);
struct quad_scope_end	*quad_scope_end();
struct quad_func_end	*quad_func_end();
struct quad_store_access *quad_store_access(const char *name, struct quad *idx, struct quad *rvalue);
struct quad_load_access	 *quad_load_access(const char *name, struct quad *lvalue, struct quad *idx);

#ifdef _MEM_AREA
struct quad_load_mem_area *quad_load_mem_area();
struct quad_store_mem_area *quad_store_mem_area();
#endif

void quad_ref(struct quad *q);
void quad_unref(struct quad *q);

#endif /* QUAD_H_ */
